import subprocess
import sys
import re
import ast
import os
import shutil
import importlib.metadata

# List of standard library modules to ignore
IGNORED_LIB_MODULES = {'os', 'enum', 'random', 'readme_version_logger'}

def get_python_version():
    return sys.version.split()[0]

def run_subprocess(command):
    try:
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if result.returncode != 0:
            print(f"Error running command {' '.join(command)}: {result.stderr}")
            return []
        return result.stdout.splitlines()
    except Exception as e:
        print(f"Exception running command {' '.join(command)}: {e}")
        return []

def get_installed_packages():
    packages = {}

    # Check if mamba is available
    if shutil.which('mamba'):
        # Get packages from mamba
        mamba_lines = run_subprocess(['mamba', 'list'])
        for line in mamba_lines:
            if line.startswith('#'):
                continue
            parts = re.split(r'\s+', line)
            if len(parts) >= 2:
                package = parts[0]
                version = parts[1]
                packages[package] = version

    # Get packages from pip
    pip_lines = run_subprocess([sys.executable, '-m', 'pip', 'freeze'])
    for line in pip_lines:
        if '==' in line:
            pkg, version = line.split('==')
            if pkg not in packages:  # Do not overwrite mamba-installed packages
                packages[pkg] = version

    return packages

def get_imported_packages(script_path):
    with open(script_path, 'r') as file:
        tree = ast.parse(file.read(), filename=script_path)
    
    imports = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                imports.add(alias.name.split('.')[0])
        elif isinstance(node, ast.ImportFrom):
            imports.add(node.module.split('.')[0])
    return list(imports)

def get_specific_packages_versions(imported_packages, installed_packages):
    specific_versions = {}
    for package in imported_packages:
        if package in IGNORED_LIB_MODULES:
            continue
        if package in installed_packages:
            specific_versions[package] = installed_packages[package]
        else:
            try:
                version = importlib.metadata.version(package)
                specific_versions[package] = f"standard library ({version})"
            except importlib.metadata.PackageNotFoundError:
                specific_versions[package] = "version not found"
    return specific_versions

def update_setup_and_run(script_path):
    python_version = get_python_version()
    installed_packages = get_installed_packages()
    imported_packages = get_imported_packages(script_path)
    specific_versions = get_specific_packages_versions(imported_packages, installed_packages)

    # Handle case where README.md does not exist
    if os.path.exists('README.md'):
        with open('README.md', 'r') as file:
            readme_content = file.read()
    else:
        readme_content = ""

    # Define the new setup and run instructions
    packages_with_versions_string = ', '.join([f"{pkg}={ver}" for pkg, ver in specific_versions.items()])
    new_setup_instructions = f"""
## Setup and Run Instructions
NOTE: This section is autogenerated by {os.path.basename(__file__)}, manual updates will be overwritten

This will guide you through the process of setting up a Mamba environment and running the provided Python code to see it in action. It uses the last known working versions of Python and packages used.

### Prerequisites

Ensure you have [Mamba](https://mamba.readthedocs.io/en/latest/installation.html) installed on your system. Mamba is a fast, cross-platform package manager.

### Steps

1. **Create a Mamba Environment**
   
   Open your terminal and execute the following commands:

   ```bash
   mamba create -n myenv python={python_version} -y
   mamba activate myenv

2. **Install Necessary Packages**

    ```bash
    mamba install {packages_with_versions_string}

3. **Run the Script**

    Ensure you are in your project directory (the directory containing your script) and run:

    ```bash
    python {script_path}

    Or click 'run' in your IDE of choice.

    <!-- END SETUP AND RUN INSTRUCTIONS -->
"""

    # Replace or append the setup instructions
    setup_instructions_marker = "## Setup and Run Instructions"
    setup_instructions_closing_marker = "<!-- END SETUP AND RUN INSTRUCTIONS -->"
    if setup_instructions_marker in readme_content:
        setup_instructions_pattern = re.compile(rf"{re.escape(setup_instructions_marker)}.*?{re.escape(setup_instructions_closing_marker)}", re.DOTALL)
        readme_content = setup_instructions_pattern.sub(new_setup_instructions, readme_content)
    else:
        readme_content += '\n' + new_setup_instructions

    with open('README.md', 'w') as file:
        file.write(readme_content)


